@@ -19,12 +19,8 @@
 //! [`parsing`]: parsing/index.html
 //! [`highlighting`]: highlighting/index.html
 
-#![doc(html_root_url = "https://docs.rs/syntect/5.0.0")]
+#![doc(html_root_url = "https://docs.rs/syntect/5.1.0")]
 
-#[macro_use]
-extern crate lazy_static;
-#[macro_use]
-extern crate serde_derive;
 #[cfg(test)]
 #[macro_use]
 extern crate pretty_assertions;
