@@ -1,3 +1,15 @@
+// Suppression of a false positive clippy lint. Upstream issue:
+//
+//   mutable_key_type false positive for raw pointers
+//   https://github.com/rust-lang/rust-clippy/issues/6745
+//
+// We use `*const MatchPattern` as key in our `SearchCache` hash map.
+// Clippy thinks this is a problem since `MatchPattern` has interior mutability
+// via `MatchPattern::regex::regex` which is an `AtomicLazyCell`.
+// But raw pointers are hashed via the pointer itself, not what is pointed to.
+// See https://github.com/rust-lang/rust/blob/1.54.0/library/core/src/hash/mod.rs#L717-L725
+#![allow(clippy::mutable_key_type)]
+
 use super::syntax_definition::*;
 use super::scope::*;
 use super::regex::Region;
@@ -7,6 +19,23 @@ use std::i32;
 use std::hash::BuildHasherDefault;
 use fnv::FnvHasher;
 use crate::parsing::syntax_set::{SyntaxSet, SyntaxReference};
+use crate::parsing::syntax_definition::ContextId;
+
+/// Errors that can occur while parsing.
+#[derive(Debug, thiserror::Error)]
+#[non_exhaustive]
+pub enum ParsingError {
+    #[error("Somehow main context was popped from the stack")]
+    MissingMainContext,
+    /// A context is missing. Usually caused by a syntax referencing a another
+    /// syntax that is not known to syntect. See e.g. <https://github.com/trishume/syntect/issues/421>
+    #[error("Missing context with ID '{0:?}'")]
+    MissingContext(ContextId),
+    #[error("Bad index to match_at: {0}")]
+    BadMatchIndex(usize),
+    #[error("Tried to use a ContextReference that has not bee resolved yet: {0:?}")]
+    UnresolvedContextReference(ContextReference),
+}
 
 /// Keeps the current parser state (the internal syntax interpreter stack) between lines of parsing.
 ///
@@ -152,7 +181,7 @@ impl ParseState {
     /// main context of the syntax
     pub fn new(syntax: &SyntaxReference) -> ParseState {
         let start_state = StateLevel {
-            context: syntax.contexts["__start"],
+            context: syntax.context_ids()["__start"],
             prototypes: Vec::new(),
             captures: None,
         };
@@ -182,15 +211,20 @@ impl ParseState {
     /// [`ScopeStack::apply`]: struct.ScopeStack.html#method.apply
     /// [`SyntaxSet`]: struct.SyntaxSet.html
     /// [`ParseState`]: struct.ParseState.html
-    pub fn parse_line(&mut self, line: &str, syntax_set: &SyntaxSet) -> Vec<(usize, ScopeStackOp)> {
-        assert!(!self.stack.is_empty(),
-                "Somehow main context was popped from the stack");
+    pub fn parse_line(
+        &mut self,
+        line: &str,
+        syntax_set: &SyntaxSet,
+    ) -> Result<Vec<(usize, ScopeStackOp)>, ParsingError> {
+        if self.stack.is_empty() {
+            return Err(ParsingError::MissingMainContext)
+        }
         let mut match_start = 0;
         let mut res = Vec::new();
 
         if self.first_line {
             let cur_level = &self.stack[self.stack.len() - 1];
-            let context = syntax_set.get_context(&cur_level.context);
+            let context = syntax_set.get_context(&cur_level.context)?;
             if !context.meta_content_scope.is_empty() {
                 res.push((0, ScopeStackOp::Push(context.meta_content_scope[0])));
             }
@@ -211,11 +245,12 @@ impl ParseState {
             &mut regions,
             &mut non_consuming_push_at,
             &mut res
-        ) {}
+        )? {}
 
-        res
+        Ok(res)
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn parse_next_token(
         &mut self,
         line: &str,
@@ -225,7 +260,7 @@ impl ParseState {
         regions: &mut Region,
         non_consuming_push_at: &mut (usize, usize),
         ops: &mut Vec<(usize, ScopeStackOp)>,
-    ) -> bool {
+    ) -> Result<bool, ParsingError> {
         let check_pop_loop = {
             let (pos, stack_depth) = *non_consuming_push_at;
             pos == *start && stack_depth == self.stack.len()
@@ -236,7 +271,7 @@ impl ParseState {
             self.proto_starts.pop();
         }
 
-        let best_match = self.find_best_match(line, *start, syntax_set, search_cache, regions, check_pop_loop);
+        let best_match = self.find_best_match(line, *start, syntax_set, search_cache, regions, check_pop_loop)?;
 
         if let Some(reg_match) = best_match {
             if reg_match.would_loop {
@@ -255,11 +290,11 @@ impl ParseState {
                 // unicode characters can be more than 1 byte.
                 if let Some((i, _)) = line[*start..].char_indices().nth(1) {
                     *start += i;
-                    return true;
+                    return Ok(true);
                 } else {
                     // End of line, no character to advance and no point trying
                     // any more patterns.
-                    return false;
+                    return Ok(false);
                 }
             }
 
@@ -272,7 +307,7 @@ impl ParseState {
                 // check the next "pop" for loops. Otherwise leave the state,
                 // e.g. non-consuming "set" could also result in a loop.
                 let context = reg_match.context;
-                let match_pattern = context.match_at(reg_match.pat_index);
+                let match_pattern = context.match_at(reg_match.pat_index)?;
                 if let MatchOperation::Push(_) = match_pattern.operation {
                     *non_consuming_push_at = (match_end, self.stack.len() + 1);
                 }
@@ -288,13 +323,13 @@ impl ParseState {
 
             let level_context = {
                 let id = &self.stack[self.stack.len() - 1].context;
-                syntax_set.get_context(id)
+                syntax_set.get_context(id)?
             };
-            self.exec_pattern(line, &reg_match, level_context, syntax_set, ops);
+            self.exec_pattern(line, &reg_match, level_context, syntax_set, ops)?;
 
-            true
+            Ok(true)
         } else {
-            false
+            Ok(false)
         }
     }
 
@@ -306,9 +341,9 @@ impl ParseState {
         search_cache: &mut SearchCache,
         regions: &mut Region,
         check_pop_loop: bool,
-    ) -> Option<RegexMatch<'a>> {
+    ) -> Result<Option<RegexMatch<'a>>, ParsingError> {
         let cur_level = &self.stack[self.stack.len() - 1];
-        let context = syntax_set.get_context(&cur_level.context);
+        let context = syntax_set.get_context(&cur_level.context)?;
         let prototype = if let Some(ref p) = context.prototype {
             Some(p)
         } else {
@@ -333,8 +368,8 @@ impl ParseState {
         let mut pop_would_loop = false;
 
         for (from_with_proto, ctx, captures) in context_chain {
-            for (pat_context, pat_index) in context_iter(syntax_set, syntax_set.get_context(ctx)) {
-                let match_pat = pat_context.match_at(pat_index);
+            for (pat_context, pat_index) in context_iter(syntax_set, syntax_set.get_context(ctx)?) {
+                let match_pat = pat_context.match_at(pat_index)?;
 
                 if let Some(match_region) = self.search(
                     line, start, match_pat, captures, search_cache, regions
@@ -353,10 +388,9 @@ impl ParseState {
                         min_start = match_start;
 
                         let consuming = match_end > start;
-                        pop_would_loop = check_pop_loop && !consuming && match match_pat.operation {
-                            MatchOperation::Pop => true,
-                            _ => false,
-                        };
+                        pop_would_loop = check_pop_loop
+                            && !consuming
+                            && matches!(match_pat.operation, MatchOperation::Pop);
 
                         best_match = Some(RegexMatch {
                             regions: match_region,
@@ -369,13 +403,13 @@ impl ParseState {
                         if match_start == start && !pop_would_loop {
                             // We're not gonna find a better match after this,
                             // so as an optimization we can stop matching now.
-                            return best_match;
+                            return Ok(best_match);
                         }
                     }
                 }
             }
         }
-        best_match
+        Ok(best_match)
     }
 
     fn search(&self,
@@ -403,15 +437,18 @@ impl ParseState {
             }
         }
 
-        let (matched, can_cache) = if match_pat.has_captures && captures.is_some() {
-            let &(ref region, ref s) = captures.unwrap();
-            let regex = match_pat.regex_with_refs(region, s);
-            let matched = regex.search(line, start, line.len(), Some(regions));
-            (matched, false)
-        } else {
-            let regex = match_pat.regex();
-            let matched = regex.search(line, start, line.len(), Some(regions));
-            (matched, true)
+        let (matched, can_cache) = match (match_pat.has_captures, captures) {
+            (true, Some(captures)) => {
+                let &(ref region, ref s) = captures;
+                let regex = match_pat.regex_with_refs(region, s);
+                let matched = regex.search(line, start, line.len(), Some(regions));
+                (matched, false)
+            }
+            _ => {
+                let regex = match_pat.regex();
+                let matched = regex.search(line, start, line.len(), Some(regions));
+                (matched, true)
+            }
         };
 
         if matched {
@@ -442,13 +479,13 @@ impl ParseState {
         level_context: &'a Context,
         syntax_set: &'a SyntaxSet,
         ops: &mut Vec<(usize, ScopeStackOp)>,
-    ) -> bool {
+    ) -> Result<bool, ParsingError> {
         let (match_start, match_end) = reg_match.regions.pos(0).unwrap();
         let context = reg_match.context;
-        let pat = context.match_at(reg_match.pat_index);
+        let pat = context.match_at(reg_match.pat_index)?;
         // println!("running pattern {:?} on '{}' at {}, operation {:?}", pat.regex_str, line, match_start, pat.operation);
 
-        self.push_meta_ops(true, match_start, level_context, &pat.operation, syntax_set, ops);
+        self.push_meta_ops(true, match_start, level_context, &pat.operation, syntax_set, ops)?;
         for s in &pat.scope {
             // println!("pushing {:?} at {}", s, match_start);
             ops.push((match_start, ScopeStackOp::Push(*s)));
@@ -481,7 +518,7 @@ impl ParseState {
             // println!("popping at {}", match_end);
             ops.push((match_end, ScopeStackOp::Pop(pat.scope.len())));
         }
-        self.push_meta_ops(false, match_end, &*level_context, &pat.operation, syntax_set, ops);
+        self.push_meta_ops(false, match_end, &*level_context, &pat.operation, syntax_set, ops)?;
 
         self.perform_op(line, &reg_match.regions, pat, syntax_set)
     }
@@ -494,7 +531,7 @@ impl ParseState {
         match_op: &MatchOperation,
         syntax_set: &'a SyntaxSet,
         ops: &mut Vec<(usize, ScopeStackOp)>,
-    ) {
+    ) -> Result<(), ParsingError>{
         // println!("metas ops for {:?}, initial: {}",
         //          match_op,
         //          initial);
@@ -512,7 +549,7 @@ impl ParseState {
 
                 // cleared scopes are restored after the scopes from match pattern that invoked the pop are applied
                 if !initial && cur_context.clear_scopes != None {
-                    ops.push((index, ScopeStackOp::Restore));
+                    ops.push((index, ScopeStackOp::Restore))
                 }
             },
             // for some reason the ST3 behaviour of set is convoluted and is inconsistent with the docs and other ops
@@ -521,10 +558,7 @@ impl ParseState {
             // - the interaction with meta scopes means that the token has the meta scopes of both the current scope and the new scope.
             MatchOperation::Push(ref context_refs) |
             MatchOperation::Set(ref context_refs) => {
-                let is_set = match *match_op {
-                    MatchOperation::Set(_) => true,
-                    _ => false
-                };
+                let is_set = matches!(*match_op, MatchOperation::Set(_));
                 // a match pattern that "set"s keeps the meta_content_scope and meta_scope from the previous context
                 if initial {
                     if is_set && cur_context.clear_scopes != None {
@@ -533,7 +567,7 @@ impl ParseState {
                     }
                     // add each context's meta scope
                     for r in context_refs.iter() {
-                        let ctx = r.resolve(syntax_set);
+                        let ctx = r.resolve(syntax_set)?;
 
                         if !is_set {
                             if let Some(clear_amount) = ctx.clear_scopes {
@@ -547,14 +581,14 @@ impl ParseState {
                     }
                 } else {
                     let repush = (is_set && (!cur_context.meta_scope.is_empty() || !cur_context.meta_content_scope.is_empty())) || context_refs.iter().any(|r| {
-                        let ctx = r.resolve(syntax_set);
+                        let ctx = r.resolve(syntax_set).unwrap();
 
                         !ctx.meta_content_scope.is_empty() || (ctx.clear_scopes.is_some() && is_set)
                     });
                     if repush {
                         // remove previously pushed meta scopes, so that meta content scopes will be applied in the correct order
                         let mut num_to_pop : usize = context_refs.iter().map(|r| {
-                            let ctx = r.resolve(syntax_set);
+                            let ctx = r.resolve(syntax_set).unwrap();
                             ctx.meta_scope.len()
                         }).sum();
 
@@ -570,7 +604,7 @@ impl ParseState {
 
                         // now we push meta scope and meta context scope for each context pushed
                         for r in context_refs {
-                            let ctx = r.resolve(syntax_set);
+                            let ctx = r.resolve(syntax_set)?;
 
                             // for some reason, contrary to my reading of the docs, set does this after the token
                             if is_set {
@@ -591,6 +625,8 @@ impl ParseState {
             },
             MatchOperation::None => (),
         }
+
+        Ok(())
     }
 
     /// Returns true if the stack was changed
@@ -600,7 +636,7 @@ impl ParseState {
         regions: &Region,
         pat: &MatchPattern,
         syntax_set: &SyntaxSet
-    ) -> bool {
+    ) -> Result<bool, ParsingError> {
         let (ctx_refs, old_proto_ids) = match pat.operation {
             MatchOperation::Push(ref ctx_refs) => (ctx_refs, None),
             MatchOperation::Set(ref ctx_refs) => {
@@ -611,15 +647,15 @@ impl ParseState {
             }
             MatchOperation::Pop => {
                 self.stack.pop();
-                return true;
+                return Ok(true);
             }
-            MatchOperation::None => return false,
+            MatchOperation::None => return Ok(false),
         };
         for (i, r) in ctx_refs.iter().enumerate() {
             let mut proto_ids = if i == 0 {
                 // it is only necessary to preserve the old prototypes
                 // at the first stack frame pushed
-                old_proto_ids.clone().unwrap_or_else(|| Vec::new())
+                old_proto_ids.clone().unwrap_or_else(Vec::new)
             } else {
                 Vec::new()
             };
@@ -630,15 +666,15 @@ impl ParseState {
                 // referred to as the "target" of the push by sublimehq - see
                 // https://forum.sublimetext.com/t/dev-build-3111/19240/17 for more info
                 if let Some(ref p) = pat.with_prototype {
-                    proto_ids.push(p.id().clone());
+                    proto_ids.push(p.id()?);
                 }
             }
-            let context_id = r.id();
-            let context = syntax_set.get_context(&context_id);
+            let context_id = r.id()?;
+            let context = syntax_set.get_context(&context_id)?;
             let captures = {
                 let mut uses_backrefs = context.uses_backrefs;
                 if !proto_ids.is_empty() {
-                    uses_backrefs = uses_backrefs || proto_ids.iter().any(|id| syntax_set.get_context(id).uses_backrefs);
+                    uses_backrefs = uses_backrefs || proto_ids.iter().any(|id| syntax_set.get_context(id).unwrap().uses_backrefs);
                 }
                 if uses_backrefs {
                     Some((regions.clone(), line.to_owned()))
@@ -652,7 +688,7 @@ impl ParseState {
                 captures,
             });
         }
-        true
+        Ok(true)
     }
 }
 
@@ -743,7 +779,7 @@ mod tests {
 
         let mut stack = ScopeStack::new();
         for &(_, ref op) in ops.iter() {
-            stack.apply(op);
+            stack.apply(op).expect("#[cfg(test)]");
         }
         assert_eq!(stack, test_stack);
     }
@@ -871,7 +907,7 @@ mod tests {
             "<source.test>, <example.meta-scope.after-clear-scopes.example>, <example.pops-clear-scopes.example>",
             "<source.test>, <string.quoted.single.example>, <constant.character.escape.example>",
         ];
-        expect_scope_stacks(&line, &expect, TEST_SYNTAX);
+        expect_scope_stacks(line, &expect, TEST_SYNTAX);
     }
 
     #[test]
@@ -882,7 +918,7 @@ mod tests {
             "<example.meta-scope.after-clear-scopes.example>, <example.pops-clear-scopes.example>",
             "<source.test>, <string.quoted.single.example>, <constant.character.escape.example>",
         ];
-        expect_scope_stacks(&line, &expect, TEST_SYNTAX);
+        expect_scope_stacks(line, &expect, TEST_SYNTAX);
     }
 
     #[test]
@@ -894,7 +930,7 @@ mod tests {
             "<source.test>, <example.meta-scope.after-clear-scopes.example>, <example.pops-clear-scopes.example>",
             "<source.test>, <string.quoted.single.example>, <constant.character.escape.example>",
         ];
-        expect_scope_stacks(&line, &expect, TEST_SYNTAX);
+        expect_scope_stacks(line, &expect, TEST_SYNTAX);
     }
 
     #[test]
@@ -903,7 +939,7 @@ mod tests {
         let expect = [
             "<source.test>, <constant.numeric.test>",
         ];
-        expect_scope_stacks(&line, &expect, TEST_SYNTAX);
+        expect_scope_stacks(line, &expect, TEST_SYNTAX);
     }
 
     #[test]
@@ -916,7 +952,7 @@ mod tests {
             "<source.test>, <test>, <string.unquoted.test>",
             "<source.test>, <test>, <keyword.control.test>",
         ];
-        expect_scope_stacks(&line, &expect, TEST_SYNTAX);
+        expect_scope_stacks(line, &expect, TEST_SYNTAX);
     }
 
     #[test]
@@ -937,7 +973,7 @@ contexts:
 
         let line = "foo!";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -955,7 +991,7 @@ contexts:
 
         let line = "foo";
         let expect = ["<source.test>, <foo.end>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
 
         let line = "foofoofoo";
         let expect = [
@@ -963,7 +999,7 @@ contexts:
             "<source.test>, <foo.any>",
             "<source.test>, <foo.end>",
         ];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -983,7 +1019,7 @@ contexts:
             "<top-level.test>",
         ];
 
-        expect_scope_stacks_with_syntax(&line1, &expect1, syntax.clone());
+        expect_scope_stacks_with_syntax(line1, &expect1, syntax.clone());
 
         let line2 = ">abctest</style>foobar";
         let expect2 = [
@@ -991,7 +1027,7 @@ contexts:
             "<source.css.embedded.html>, <test.embedded>",
             "<top-level.test>",
         ];
-        expect_scope_stacks_with_syntax(&line2, &expect2, syntax.clone());
+        expect_scope_stacks_with_syntax(line2, &expect2, syntax);
     }
 
     #[test]
@@ -1017,7 +1053,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.matched>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1046,7 +1082,7 @@ contexts:
 
         let line = "test";
         let expect = ["<source.test>, <test.matched>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1074,7 +1110,7 @@ contexts:
 
         let line = "test";
         let expect = ["<source.test>, <test.matched>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1103,7 +1139,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.matched>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1125,7 +1161,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.matched>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1153,7 +1189,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1188,7 +1224,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1216,7 +1252,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1248,7 +1284,7 @@ contexts:
 
         let line = "foo::bar::* xxx";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1281,7 +1317,7 @@ contexts:
 
         let line = "hello";
         let expect = ["<source.test>, <test.good>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1308,7 +1344,7 @@ contexts:
     - match: ''
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax, true, None).unwrap();
         expect_scope_stacks_with_syntax("/** * */", &["<comment.block.documentation.javadoc>, <punctuation.definition.comment.begin.javadoc>", "<comment.block.documentation.javadoc>, <text.html.javadoc>, <punctuation.definition.comment.javadoc>", "<comment.block.documentation.javadoc>, <punctuation.definition.comment.end.javadoc>"], syntax);
     }
 
@@ -1392,13 +1428,13 @@ contexts:
             "a1b2c3d4e5",
             &[
                 "<a>", "<1>", "<b>", "<2>", "<c>", "<3>", "<d>", "<4>", "<e>", "<5>"
-            ], SyntaxDefinition::load_from_str(&syntax, true, None).unwrap()
+            ], SyntaxDefinition::load_from_str(syntax, true, None).unwrap()
         );
         expect_scope_stacks_with_syntax(
             "5cfcecbedcdea",
             &[
                 "<5>", "<cwith>", "<f>", "<e>", "<b>", "<d>", "<cwithout>", "<a>"
-            ], SyntaxDefinition::load_from_str(&syntax, true, None).unwrap()
+            ], SyntaxDefinition::load_from_str(syntax, true, None).unwrap()
         );
     }
 
@@ -1419,7 +1455,7 @@ contexts:
             1: keyword
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax, true, None).unwrap();
         expect_scope_stacks_with_syntax("testfoo", &["<test>", /*"<ignored>",*/ "<f>", "<keyword>"], syntax);
     }
 
@@ -1448,7 +1484,7 @@ contexts:
           scope: '1'
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax_yamlstr, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax_yamlstr, true, None).unwrap();
         expect_scope_stacks_with_syntax("abc12", &["<1>", "<2>"], syntax);
     }
 
@@ -1495,11 +1531,11 @@ contexts:
       pop: true
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax_yamlstr, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax_yamlstr, true, None).unwrap();
         expect_scope_stacks_with_syntax("ab12", &["<1>", "<2>"], syntax.clone());
         expect_scope_stacks_with_syntax("abc12", &["<1>", "<digit2>"], syntax.clone());
         expect_scope_stacks_with_syntax("abcd12", &["<1>", "<digit2>"], syntax.clone());
-        expect_scope_stacks_with_syntax("abcde12", &["<digit1>", "<digit2>"], syntax.clone());
+        expect_scope_stacks_with_syntax("abcde12", &["<digit1>", "<digit2>"], syntax);
     }
 
     #[test]
@@ -1528,7 +1564,7 @@ contexts:
           pop: true
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax_yamlstr, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax_yamlstr, true, None).unwrap();
         expect_scope_stacks_with_syntax("aa", &["<a>", "<1>"], syntax.clone());
         expect_scope_stacks_with_syntax("abcdb", &["<a>", "<b>", "<c>", "<d>", "<1>"], syntax);
     }
@@ -1562,7 +1598,7 @@ contexts:
           pop: true
 "#;
 
-        let syntax = SyntaxDefinition::load_from_str(&syntax_yamlstr, true, None).unwrap();
+        let syntax = SyntaxDefinition::load_from_str(syntax_yamlstr, true, None).unwrap();
         expect_scope_stacks_with_syntax("a--", &["<a>", "<2>"], syntax.clone());
         // it seems that when ST encounters a non existing pop backreference, it just pops back to the with_prototype's original parent context - i.e. cdb is unscoped
         // TODO: it would be useful to have syntest functionality available here for easier testing and clarity
@@ -1582,7 +1618,7 @@ contexts:
 
         let line = "foo";
         let expect = ["<source.test>, <foo.newline>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1598,7 +1634,7 @@ contexts:
 
         let line = "foo";
         let expect = ["<source.test>, <foo.newline>"];
-        expect_scope_stacks(&line, &expect, syntax);
+        expect_scope_stacks(line, &expect, syntax);
     }
 
     #[test]
@@ -1621,7 +1657,7 @@ contexts:
       scope: other
 "#;
 
-        let syntax_newlines = SyntaxDefinition::load_from_str(&syntax, true, None).unwrap();
+        let syntax_newlines = SyntaxDefinition::load_from_str(syntax, true, None).unwrap();
         let syntax_set = link(syntax_newlines);
 
         let mut state = ParseState::new(&syntax_set.syntaxes()[0]);
@@ -1659,7 +1695,7 @@ contexts:
       scope: comment.line.double-slash
 "#;
 
-        let syntax_newlines = SyntaxDefinition::load_from_str(&syntax, true, None).unwrap();
+        let syntax_newlines = SyntaxDefinition::load_from_str(syntax, true, None).unwrap();
         let syntax_set = link(syntax_newlines);
 
         let mut state = ParseState::new(&syntax_set.syntaxes()[0]);
@@ -1715,7 +1751,7 @@ contexts:
                       pop: true
                 "#, true, None).unwrap();
 
-        expect_scope_stacks_with_syntax(&"aa", &["<a>", "<b>"], syntax);
+        expect_scope_stacks_with_syntax("aa", &["<a>", "<b>"], syntax);
     }
     
     #[test]
@@ -1738,18 +1774,18 @@ contexts:
                       pop: true
                 "#, true, None).unwrap();
 
-        expect_scope_stacks_with_syntax(&"aa", &["<a>", "<b>"], syntax);
+        expect_scope_stacks_with_syntax("aa", &["<a>", "<b>"], syntax);
     }
 
     fn expect_scope_stacks(line_without_newline: &str, expect: &[&str], syntax: &str) {
         println!("Parsing with newlines");
         let line_with_newline = format!("{}\n", line_without_newline);
-        let syntax_newlines = SyntaxDefinition::load_from_str(&syntax, true, None).unwrap();
+        let syntax_newlines = SyntaxDefinition::load_from_str(syntax, true, None).unwrap();
         expect_scope_stacks_with_syntax(&line_with_newline, expect, syntax_newlines);
 
         println!("Parsing without newlines");
-        let syntax_nonewlines = SyntaxDefinition::load_from_str(&syntax, false, None).unwrap();
-        expect_scope_stacks_with_syntax(&line_without_newline, expect, syntax_nonewlines);
+        let syntax_nonewlines = SyntaxDefinition::load_from_str(syntax, false, None).unwrap();
+        expect_scope_stacks_with_syntax(line_without_newline, expect, syntax_nonewlines);
     }
 
     fn expect_scope_stacks_with_syntax(line: &str, expect: &[&str], syntax: SyntaxDefinition) {
@@ -1771,7 +1807,7 @@ contexts:
                 }
             }
         }
-        if let Some(missing) = expect.iter().filter(|e| !criteria_met.contains(&e)).next() {
+        if let Some(missing) = expect.iter().find(|e| !criteria_met.contains(e)) {
             panic!("expected scope stack '{}' missing", missing);
         }
     }
@@ -1791,7 +1827,7 @@ contexts:
     }
 
     fn ops(state: &mut ParseState, line: &str, syntax_set: &SyntaxSet) -> Vec<(usize, ScopeStackOp)> {
-        let ops = state.parse_line(line, syntax_set);
+        let ops = state.parse_line(line, syntax_set).expect("#[cfg(test)]");
         debug_print_ops(line, &ops);
         ops
     }
@@ -1800,7 +1836,7 @@ contexts:
         let mut states = Vec::new();
         let mut stack = ScopeStack::new();
         for &(_, ref op) in ops.iter() {
-            stack.apply(op);
+            stack.apply(op).expect("#[cfg(test)]");
             let scopes: Vec<String> = stack.as_slice().iter().map(|s| format!("{:?}", s)).collect();
             let stack_str = scopes.join(", ");
             states.push(stack_str);
