@@ -1,3 +1,4 @@
+use super::ParsingError;
 use super::syntax_definition::*;
 use super::scope::*;
 
@@ -9,17 +10,14 @@ use super::super::LoadingError;
 
 use std::collections::{HashMap, HashSet, BTreeSet};
 use std::path::Path;
-#[cfg(feature = "yaml-load")]
-use walkdir::WalkDir;
-#[cfg(feature = "yaml-load")]
-use std::io::Read;
 use std::io::{self, BufRead, BufReader};
 use std::fs::File;
 use std::mem;
 
-use lazycell::AtomicLazyCell;
 use super::regex::Regex;
 use crate::parsing::syntax_definition::ContextId;
+use once_cell::sync::OnceCell;
+use serde::{Deserialize, Serialize};
 
 /// A syntax set holds multiple syntaxes that have been linked together.
 ///
@@ -33,12 +31,11 @@ use crate::parsing::syntax_definition::ContextId;
 #[derive(Debug, Serialize, Deserialize)]
 pub struct SyntaxSet {
     syntaxes: Vec<SyntaxReference>,
-    contexts: Vec<Context>,
     /// Stores the syntax index for every path that was loaded
     path_syntaxes: Vec<(String, usize)>,
 
-    #[serde(skip_serializing, skip_deserializing, default = "AtomicLazyCell::new")]
-    first_line_cache: AtomicLazyCell<FirstLineCache>,
+    #[serde(skip_serializing, skip_deserializing, default = "OnceCell::new")]
+    first_line_cache: OnceCell<FirstLineCache>,
     /// Metadata, e.g. indent and commenting information.
     ///
     /// NOTE: if serializing, you should handle metadata manually; that is, you should serialize and
@@ -60,8 +57,17 @@ pub struct SyntaxReference {
     pub hidden: bool,
     #[serde(serialize_with = "ordered_map")]
     pub variables: HashMap<String, String>,
+    #[serde(skip)]
+    pub(crate) lazy_contexts: OnceCell<LazyContexts>,
+    pub(crate) serialized_lazy_contexts: Vec<u8>,
+}
+
+/// The lazy-loaded parts of a [`SyntaxReference`].
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub(crate) struct LazyContexts {
     #[serde(serialize_with = "ordered_map")]
-    pub(crate) contexts: HashMap<String, ContextId>,
+    pub(crate) context_ids: HashMap<String, ContextId>,
+    pub(crate) contexts: Vec<Context>,
 }
 
 /// A syntax set builder is used for loading syntax definitions from the file
@@ -91,27 +97,23 @@ pub struct SyntaxSetBuilder {
 fn load_syntax_file(p: &Path,
                     lines_include_newline: bool)
                     -> Result<SyntaxDefinition, LoadingError> {
-    let mut f = File::open(p)?;
-    let mut s = String::new();
-    f.read_to_string(&mut s)?;
+    let s = std::fs::read_to_string(p)?;
 
-    Ok(
-        SyntaxDefinition::load_from_str(
-            &s,
-            lines_include_newline,
-            p.file_stem().and_then(|x| x.to_str())
-        ).map_err(|e| LoadingError::ParseSyntax(e, Some(format!("{}", p.display()))))?
+    SyntaxDefinition::load_from_str(
+        &s,
+        lines_include_newline,
+        p.file_stem().and_then(|x| x.to_str()),
     )
+    .map_err(|e| LoadingError::ParseSyntax(e, format!("{}", p.display())))
 }
 
 impl Clone for SyntaxSet {
     fn clone(&self) -> SyntaxSet {
         SyntaxSet {
             syntaxes: self.syntaxes.clone(),
-            contexts: self.contexts.clone(),
             path_syntaxes: self.path_syntaxes.clone(),
             // Will need to be re-initialized
-            first_line_cache: AtomicLazyCell::new(),
+            first_line_cache: OnceCell::new(),
             #[cfg(feature = "metadata")]
             metadata: self.metadata.clone(),
         }
@@ -122,16 +124,14 @@ impl Default for SyntaxSet {
     fn default() -> Self {
         SyntaxSet {
             syntaxes: Vec::new(),
-            contexts: Vec::new(),
             path_syntaxes: Vec::new(),
-            first_line_cache: AtomicLazyCell::new(),
+            first_line_cache: OnceCell::new(),
             #[cfg(feature = "metadata")]
             metadata: Metadata::default(),
         }
     }
 }
 
-
 impl SyntaxSet {
     pub fn new() -> SyntaxSet {
         SyntaxSet::default()
@@ -183,7 +183,7 @@ impl SyntaxSet {
     }
 
     pub fn find_syntax_by_extension<'a>(&'a self, extension: &str) -> Option<&'a SyntaxReference> {
-        self.syntaxes.iter().rev().find(|&s| s.file_extensions.iter().any(|e| e == extension))
+        self.syntaxes.iter().rev().find(|&s| s.file_extensions.iter().any(|e| e.eq_ignore_ascii_case(extension)))
     }
 
     /// Searches for a syntax first by extension and then by case-insensitive name
@@ -223,7 +223,7 @@ impl SyntaxSet {
     // however, if a syntax name is provided without a folder, make sure we don't accidentally match the end of a different syntax definition's name - by checking a / comes before it or it is the full path
     pub fn find_syntax_by_path<'a>(&'a self, path: &str) -> Option<&'a SyntaxReference> {
         let mut slash_path = "/".to_string();
-        slash_path.push_str(&path);
+        slash_path.push_str(path);
         self.path_syntaxes.iter().rev().find(|t| t.0.ends_with(&slash_path) || t.0 == path).map(|&(_,i)| &self.syntaxes[i])
     }
 
@@ -297,13 +297,15 @@ impl SyntaxSet {
     /// in the set, but not the other way around.
     pub fn into_builder(self) -> SyntaxSetBuilder {
         #[cfg(feature = "metadata")]
-        let SyntaxSet { syntaxes, contexts, path_syntaxes, metadata, .. } = self;
+        let SyntaxSet { syntaxes, path_syntaxes, metadata, .. } = self;
         #[cfg(not(feature = "metadata"))]
-        let SyntaxSet { syntaxes, contexts, path_syntaxes, .. } = self;
+        let SyntaxSet { syntaxes, path_syntaxes, .. } = self;
 
-        let mut context_map = HashMap::with_capacity(contexts.len());
-        for (i, context) in contexts.into_iter().enumerate() {
-            context_map.insert(i, context);
+        let mut context_map = HashMap::new();
+        for (syntax_index, syntax) in syntaxes.iter().enumerate() {
+            for (context_index, context) in syntax.contexts().iter().enumerate() {
+                context_map.insert(ContextId { syntax_index, context_index }, context.clone());
+            }
         }
 
         let mut builder_syntaxes = Vec::with_capacity(syntaxes.len());
@@ -316,12 +318,14 @@ impl SyntaxSet {
                 first_line_match,
                 hidden,
                 variables,
-                contexts,
+                serialized_lazy_contexts,
+                ..
             } = syntax;
 
-            let mut builder_contexts = HashMap::with_capacity(contexts.len());
-            for (name, context_id) in contexts {
-                if let Some(context) = context_map.remove(&context_id.index()) {
+            let lazy_contexts = LazyContexts::deserialize(&serialized_lazy_contexts[..]);
+            let mut builder_contexts = HashMap::with_capacity(lazy_contexts.context_ids.len());
+            for (name, context_id) in lazy_contexts.context_ids {
+                if let Some(context) = context_map.remove(&context_id) {
                     builder_contexts.insert(name, context);
                 }
             }
@@ -349,27 +353,24 @@ impl SyntaxSet {
     }
 
     #[inline(always)]
-    pub(crate) fn get_context(&self, context_id: &ContextId) -> &Context {
-        &self.contexts[context_id.index()]
+    pub(crate) fn get_context(&self, context_id: &ContextId) -> Result<&Context, ParsingError> {
+        let syntax = &self
+            .syntaxes
+            .get(context_id.syntax_index)
+            .ok_or_else(|| ParsingError::MissingContext(*context_id))?;
+        syntax
+            .contexts()
+            .get(context_id.context_index)
+            .ok_or_else(|| ParsingError::MissingContext(*context_id))
     }
 
     fn first_line_cache(&self) -> &FirstLineCache {
-        if let Some(cache) = self.first_line_cache.borrow() {
-            cache
-        } else {
-            let cache = FirstLineCache::new(self.syntaxes());
-            self.first_line_cache.fill(cache).ok();
-            self.first_line_cache.borrow().unwrap()
-        }
+        self.first_line_cache
+            .get_or_init(|| FirstLineCache::new(self.syntaxes()))
     }
 
     pub fn find_unlinked_contexts(&self) -> BTreeSet<String> {
-        let SyntaxSet { syntaxes, contexts, .. } = self;
-
-        let mut context_map = HashMap::with_capacity(contexts.len());
-        for (i, context) in contexts.into_iter().enumerate() {
-            context_map.insert(i, context);
-        }
+        let SyntaxSet { syntaxes, .. } = self;
 
         let mut unlinked_contexts = BTreeSet::new();
 
@@ -377,49 +378,66 @@ impl SyntaxSet {
             let SyntaxReference {
                 name,
                 scope,
-                contexts,
                 ..
             } = syntax;
 
-            for (_, context_id) in contexts {
-                if let Some(context) = context_map.remove(&context_id.index()) {
-                    for pattern in context.patterns.iter() {
-                        let maybe_refs_to_check = match pattern {
-                            Pattern::Match(match_pat) => {
-                                match &match_pat.operation {
-                                    MatchOperation::Push(context_refs) => {
-                                        Some(context_refs)
-                                    },
-                                    MatchOperation::Set(context_refs) => {
-                                        Some(context_refs)
-                                    },
-                                    _ => None,
-                                }
-                            },
-                            _ => None,
-                        };
-
-                        for context_ref in maybe_refs_to_check.into_iter().flatten() {
-                            match context_ref {
-                                ContextReference::Direct(_) => {},
-                                _ => {
-                                    unlinked_contexts.insert(
-                                        format!(
-                                            "Syntax '{}' with scope '{}' has unresolved context reference {:?}",
-                                            name, scope, &context_ref
-                                        )
-                                    );
-                                },
-                            }
-                        }
-                    }
-                }
+            for context in syntax.contexts() {
+                Self::find_unlinked_contexts_in_context(name, scope, context, &mut unlinked_contexts);
             }
         }
         unlinked_contexts
     }
+
+    fn find_unlinked_contexts_in_context(
+        name: &str,
+        scope: &Scope,
+        context: &Context,
+        unlinked_contexts: &mut BTreeSet<String>,
+    ) {
+        for pattern in context.patterns.iter() {
+            let maybe_refs_to_check = match pattern {
+                Pattern::Match(match_pat) => match &match_pat.operation {
+                    MatchOperation::Push(context_refs) => Some(context_refs),
+                    MatchOperation::Set(context_refs) => Some(context_refs),
+                    _ => None,
+                },
+                _ => None,
+            };
+            for context_ref in maybe_refs_to_check.into_iter().flatten() {
+                match context_ref {
+                    ContextReference::Direct(_) => {}
+                    _ => {
+                        unlinked_contexts.insert(format!(
+                            "Syntax '{}' with scope '{}' has unresolved context reference {:?}",
+                            name, scope, &context_ref
+                        ));
+                    }
+                }
+            }
+        }
+    }
 }
 
+impl SyntaxReference {
+    pub(crate) fn context_ids(&self) -> &HashMap<String, ContextId> {
+        &self.lazy_contexts().context_ids
+    }
+
+    fn contexts(&self) -> &[Context] {
+        &self.lazy_contexts().contexts
+    }
+
+    fn lazy_contexts(&self) -> &LazyContexts {
+        self.lazy_contexts
+            .get_or_init(|| LazyContexts::deserialize(&self.serialized_lazy_contexts[..]))
+    }
+}
+
+impl LazyContexts {
+    fn deserialize(data: &[u8]) -> LazyContexts {
+        crate::dumps::from_reader(data).expect("data is not corrupt or out of sync with the code")
+    }
+}
 
 impl SyntaxSetBuilder {
     pub fn new() -> SyntaxSetBuilder {
@@ -467,7 +485,7 @@ impl SyntaxSetBuilder {
         folder: P,
         lines_include_newline: bool
     ) -> Result<(), LoadingError> {
-        for entry in WalkDir::new(folder).sort_by(|a, b| a.file_name().cmp(b.file_name())) {
+        for entry in crate::utils::walk_dir(folder).sort_by(|a, b| a.file_name().cmp(b.file_name())) {
             let entry = entry.map_err(LoadingError::WalkDir)?;
             if entry.path().extension().map_or(false, |e| e == "sublime-syntax") {
                 let syntax = load_syntax_file(entry.path(), lines_include_newline)?;
@@ -529,9 +547,10 @@ impl SyntaxSetBuilder {
         } = self;
 
         let mut syntaxes = Vec::with_capacity(syntax_definitions.len());
-        let mut all_contexts = Vec::new();
+        let mut all_context_ids = Vec::new();
+        let mut all_contexts = vec![Vec::new(); syntax_definitions.len()];
 
-        for syntax_definition in syntax_definitions {
+        for (syntax_index, syntax_definition) in syntax_definitions.into_iter().enumerate() {
             let SyntaxDefinition {
                 name,
                 file_extensions,
@@ -542,18 +561,18 @@ impl SyntaxSetBuilder {
                 contexts,
             } = syntax_definition;
 
-            let mut map = HashMap::new();
+            let mut context_ids = HashMap::new();
 
             let mut contexts: Vec<(String, Context)> = contexts.into_iter().collect();
             // Sort the values of the HashMap so that the contexts in the
             // resulting SyntaxSet have a deterministic order for serializing.
             // Because we're sorting by the keys which are unique, we can use
             // an unstable sort.
-            contexts.sort_unstable_by(|(name_a, _), (name_b, _)| name_a.cmp(&name_b));
+            contexts.sort_unstable_by(|(name_a, _), (name_b, _)| name_a.cmp(name_b));
             for (name, context) in contexts {
-                let index = all_contexts.len();
-                map.insert(name, ContextId::new(index));
-                all_contexts.push(context);
+                let context_index = all_contexts[syntax_index].len();
+                context_ids.insert(name, ContextId { syntax_index, context_index });
+                all_contexts[syntax_index].push(context);
             }
 
             let syntax = SyntaxReference {
@@ -563,29 +582,30 @@ impl SyntaxSetBuilder {
                 first_line_match,
                 hidden,
                 variables,
-                contexts: map,
+                lazy_contexts: OnceCell::new(),
+                serialized_lazy_contexts: Vec::new(), // initialized in the last step
             };
             syntaxes.push(syntax);
+            all_context_ids.push(context_ids);
         }
 
         let mut found_more_backref_includes = true;
-        for syntax in &syntaxes {
+        for (syntax_index, syntax) in syntaxes.iter().enumerate() {
             let mut no_prototype = HashSet::new();
-            let prototype = syntax.contexts.get("prototype");
+            let prototype = all_context_ids[syntax_index].get("prototype");
             if let Some(prototype_id) = prototype {
                 // TODO: We could do this after parsing YAML, instead of here?
-                Self::recursively_mark_no_prototype(syntax, prototype_id.index(), &all_contexts, &mut no_prototype);
+                Self::recursively_mark_no_prototype(syntax, prototype_id, &all_context_ids[syntax_index], &all_contexts, &mut no_prototype);
             }
 
-            for context_id in syntax.contexts.values() {
-                let index = context_id.index();
-                let mut context = &mut all_contexts[index];
+            for context_id in all_context_ids[syntax_index].values() {
+                let context = &mut all_contexts[context_id.syntax_index][context_id.context_index];
                 if let Some(prototype_id) = prototype {
-                    if context.meta_include_prototype && !no_prototype.contains(&index) {
+                    if context.meta_include_prototype && !no_prototype.contains(context_id) {
                         context.prototype = Some(*prototype_id);
                     }
                 }
-                Self::link_context(&mut context, syntax, &syntaxes);
+                Self::link_context(context, syntax_index, &all_context_ids, &syntaxes);
                 
                 if context.uses_backrefs {
                     found_more_backref_includes = true;
@@ -604,19 +624,17 @@ impl SyntaxSetBuilder {
             found_more_backref_includes = false;
             // find any contexts which include a context which uses backrefs
             // and mark those as using backrefs - to support nested includes
-            for context_index in 0..all_contexts.len() {
-                let context = &all_contexts[context_index];
-                if !context.uses_backrefs && context.patterns.iter().any(|pattern| {
-                    match pattern {
-                        Pattern::Include(ContextReference::Direct(id))
-                            if all_contexts[id.index()].uses_backrefs => true,
-                        _ => false,
+            for syntax_index in 0..syntaxes.len() {
+                for context_index in 0..all_contexts[syntax_index].len() {
+                    let context = &all_contexts[syntax_index][context_index];
+                    if !context.uses_backrefs && context.patterns.iter().any(|pattern| {
+                        matches!(pattern, Pattern::Include(ContextReference::Direct(id)) if all_contexts[id.syntax_index][id.context_index].uses_backrefs)
+                    }) {
+                        let context = &mut all_contexts[syntax_index][context_index];
+                        context.uses_backrefs = true;
+                        // look for contexts including this context
+                        found_more_backref_includes = true;
                     }
-                }) {
-                    let mut context = &mut all_contexts[context_index];
-                    context.uses_backrefs = true;
-                    // look for contexts including this context
-                    found_more_backref_includes = true;
                 }
             }
         }
@@ -627,11 +645,23 @@ impl SyntaxSetBuilder {
             None => raw_metadata.into(),
         };
 
+        // The combination of
+        //  * the algorithms above
+        //  * the borrow checker
+        // makes it necessary to set these up as the last step.
+        for syntax in &mut syntaxes {
+            let lazy_contexts = LazyContexts {
+                context_ids: all_context_ids.remove(0),
+                contexts: all_contexts.remove(0),
+            };
+
+            syntax.serialized_lazy_contexts = crate::dumps::dump_binary(&lazy_contexts);
+        };
+
         SyntaxSet {
             syntaxes,
-            contexts: all_contexts,
             path_syntaxes,
-            first_line_cache: AtomicLazyCell::new(),
+            first_line_cache: OnceCell::new(),
             #[cfg(feature = "metadata")]
             metadata,
         }
@@ -641,16 +671,17 @@ impl SyntaxSetBuilder {
     /// This marks them as such.
     fn recursively_mark_no_prototype(
         syntax: &SyntaxReference,
-        context_id: usize,
-        contexts: &[Context],
-        no_prototype: &mut HashSet<usize>,
+        context_id: &ContextId,
+        syntax_context_ids: &HashMap<String, ContextId>,
+        all_contexts: &[Vec<Context>],
+        no_prototype: &mut HashSet<ContextId>,
     ) {
-        let first_time = no_prototype.insert(context_id);
+        let first_time = no_prototype.insert(*context_id);
         if !first_time {
             return;
         }
 
-        for pattern in &contexts[context_id].patterns {
+        for pattern in &all_contexts[context_id.syntax_index][context_id.context_index].patterns {
             match *pattern {
                 // Apparently inline blocks also don't include the prototype when within the prototype.
                 // This is really weird, but necessary to run the YAML syntax.
@@ -664,12 +695,12 @@ impl SyntaxSetBuilder {
                         for context_ref in context_refs.iter() {
                             match context_ref {
                                 ContextReference::Inline(ref s) | ContextReference::Named(ref s) => {
-                                    if let Some(i) = syntax.contexts.get(s) {
-                                        Self::recursively_mark_no_prototype(syntax, i.index(), contexts, no_prototype);
+                                    if let Some(i) = syntax_context_ids.get(s) {
+                                        Self::recursively_mark_no_prototype(syntax, i, syntax_context_ids, all_contexts, no_prototype);
                                     }
                                 },
                                 ContextReference::Direct(ref id) => {
-                                    Self::recursively_mark_no_prototype(syntax, id.index(), contexts, no_prototype);
+                                    Self::recursively_mark_no_prototype(syntax, id, syntax_context_ids, all_contexts, no_prototype);
                                 },
                                 _ => (),
                             }
@@ -679,12 +710,12 @@ impl SyntaxSetBuilder {
                 Pattern::Include(ref reference) => {
                     match reference {
                         ContextReference::Named(ref s) => {
-                            if let Some(id) = syntax.contexts.get(s) {
-                                Self::recursively_mark_no_prototype(syntax, id.index(), contexts, no_prototype);
+                            if let Some(id) = syntax_context_ids.get(s) {
+                                Self::recursively_mark_no_prototype(syntax, id, syntax_context_ids, all_contexts, no_prototype);
                             }
                         },
                         ContextReference::Direct(ref id) => {
-                            Self::recursively_mark_no_prototype(syntax, id.index(), contexts, no_prototype);
+                            Self::recursively_mark_no_prototype(syntax, id, syntax_context_ids, all_contexts, no_prototype);
                         },
                         _ => (),
                     }
@@ -693,16 +724,26 @@ impl SyntaxSetBuilder {
         }
     }
 
-    fn link_context(context: &mut Context, syntax: &SyntaxReference, syntaxes: &[SyntaxReference]) {
+    fn link_context(
+        context: &mut Context,
+        syntax_index: usize,
+        all_context_ids: &[HashMap<String, ContextId>],
+        syntaxes: &[SyntaxReference],
+    ) {
         for pattern in &mut context.patterns {
             match *pattern {
-                Pattern::Match(ref mut match_pat) => Self::link_match_pat(match_pat, syntax, syntaxes),
-                Pattern::Include(ref mut context_ref) => Self::link_ref(context_ref, syntax, syntaxes),
+                Pattern::Match(ref mut match_pat) => Self::link_match_pat(match_pat, syntax_index, all_context_ids, syntaxes),
+                Pattern::Include(ref mut context_ref) => Self::link_ref(context_ref, syntax_index, all_context_ids, syntaxes),
             }
         }
     }
 
-    fn link_ref(context_ref: &mut ContextReference, syntax: &SyntaxReference, syntaxes: &[SyntaxReference]) {
+    fn link_ref(
+        context_ref: &mut ContextReference,
+        syntax_index: usize,
+        all_context_ids: &[HashMap<String, ContextId>],
+        syntaxes: &[SyntaxReference],
+    ) {
         // println!("{:?}", context_ref);
         use super::syntax_definition::ContextReference::*;
         let linked_context_id = match *context_ref {
@@ -711,26 +752,20 @@ impl SyntaxSetBuilder {
                 // This is being phased out anyhow, see https://github.com/sublimehq/Packages/issues/73
                 // Fixes issue #30
                 if s == "$top_level_main" {
-                    syntax.contexts.get("main")
+                    all_context_ids[syntax_index].get("main")
                 } else {
-                    syntax.contexts.get(s)
+                    all_context_ids[syntax_index].get(s)
                 }
             }
-            ByScope { scope, ref sub_context } => {
-                let context_name = sub_context.as_ref().map_or("main", |x| &**x);
-                syntaxes
-                    .iter()
-                    .rev()
-                    .find(|s| s.scope == scope)
-                    .and_then(|s| s.contexts.get(context_name))
+            ByScope { scope, ref sub_context, with_escape } => {
+                Self::with_plain_text_fallback(all_context_ids, syntaxes, with_escape, Self::find_id(sub_context, all_context_ids, syntaxes, |index_and_syntax| {
+                    index_and_syntax.1.scope == scope
+                }))
             }
-            File { ref name, ref sub_context } => {
-                let context_name = sub_context.as_ref().map_or("main", |x| &**x);
-                syntaxes
-                    .iter()
-                    .rev()
-                    .find(|s| &s.name == name)
-                    .and_then(|s| s.contexts.get(context_name))
+            File { ref name, ref sub_context, with_escape } => {
+                Self::with_plain_text_fallback(all_context_ids, syntaxes, with_escape, Self::find_id(sub_context, all_context_ids, syntaxes, |index_and_syntax| {
+                    &index_and_syntax.1.name == name
+                }))
             }
             Direct(_) => None,
         };
@@ -740,7 +775,53 @@ impl SyntaxSetBuilder {
         }
     }
 
-    fn link_match_pat(match_pat: &mut MatchPattern, syntax: &SyntaxReference, syntaxes: &[SyntaxReference]) {
+    fn with_plain_text_fallback<'a>(
+        all_context_ids: &'a [HashMap<String, ContextId>],
+        syntaxes: &'a [SyntaxReference],
+        with_escape: bool,
+        context_id: Option<&'a ContextId>,
+    ) -> Option<&'a ContextId> {
+        context_id.or_else(|| {
+            if with_escape {
+                // If we keep this reference unresolved, syntect will crash
+                // when it encounters the reference. Rather than crashing,
+                // we instead fall back to "Plain Text". This seems to be
+                // how Sublime Text behaves. It should be a safe thing to do
+                // since `embed`s always includes an `escape` to get out of
+                // the `embed`.
+                Self::find_id(
+                    &None,
+                    all_context_ids,
+                    syntaxes,
+                    |index_and_syntax| index_and_syntax.1.name == "Plain Text",
+                )
+            } else {
+                None
+            }
+        })
+    }
+
+    fn find_id<'a>(
+        sub_context: &Option<String>,
+        all_context_ids: &'a [HashMap<String, ContextId>],
+        syntaxes: &'a [SyntaxReference],
+        predicate: impl FnMut(&(usize, &SyntaxReference)) -> bool,
+    ) -> Option<&'a ContextId> {
+        let context_name = sub_context.as_ref().map_or("main", |x| &**x);
+        syntaxes
+            .iter()
+            .enumerate()
+            .rev()
+            .find(predicate)
+            .and_then(|index_and_syntax| all_context_ids[index_and_syntax.0].get(context_name))
+    }
+
+    fn link_match_pat(
+        match_pat: &mut MatchPattern,
+        syntax_index: usize,
+        all_context_ids: &[HashMap<String, ContextId>],
+        syntaxes: &[SyntaxReference],
+    ) {
         let maybe_context_refs = match match_pat.operation {
             MatchOperation::Push(ref mut context_refs) |
             MatchOperation::Set(ref mut context_refs) => Some(context_refs),
@@ -748,11 +829,11 @@ impl SyntaxSetBuilder {
         };
         if let Some(context_refs) = maybe_context_refs {
             for context_ref in context_refs.iter_mut() {
-                Self::link_ref(context_ref, syntax, syntaxes);
+                Self::link_ref(context_ref, syntax_index, all_context_ids, syntaxes);
             }
         }
         if let Some(ref mut context_ref) = match_pat.with_prototype {
-            Self::link_ref(context_ref, syntax, syntaxes);
+            Self::link_ref(context_ref, syntax_index, all_context_ids, syntaxes);
         }
     }
 }
@@ -812,6 +893,7 @@ mod tests {
         let syntax = ps.find_syntax_by_name("Ruby on Rails").unwrap();
         ps.find_syntax_plain_text();
         assert_eq!(&ps.find_syntax_by_extension("rake").unwrap().name, "Ruby");
+        assert_eq!(&ps.find_syntax_by_extension("RAKE").unwrap().name, "Ruby");
         assert_eq!(&ps.find_syntax_by_token("ruby").unwrap().name, "Ruby");
         assert_eq!(&ps.find_syntax_by_first_line("lol -*- Mode: C -*- such line").unwrap().name,
                    "C");
@@ -834,8 +916,8 @@ mod tests {
                    "Rust");
         // println!("{:#?}", syntax);
         assert_eq!(syntax.scope, rails_scope);
-        // assert!(false);
-        let main_context = ps.get_context(&syntax.contexts["main"]);
+        // unreachable!();
+        let main_context = ps.get_context(&syntax.context_ids()["main"]).expect("#[cfg(test)]");
         let count = syntax_definition::context_iter(&ps, main_context).count();
         assert_eq!(count, 109);
     }
@@ -848,13 +930,14 @@ mod tests {
             builder.add(syntax_b());
 
             let syntax_set_original = builder.build();
+            #[allow(clippy::redundant_clone)] // We want to test .clone()
             syntax_set_original.clone()
             // Note: The original syntax set is dropped
         };
 
         let syntax = cloned_syntax_set.find_syntax_by_extension("a").unwrap();
         let mut parse_state = ParseState::new(syntax);
-        let ops = parse_state.parse_line("a go_b b", &cloned_syntax_set);
+        let ops = parse_state.parse_line("a go_b b", &cloned_syntax_set).expect("#[cfg(test)]");
         let expected = (7, ScopeStackOp::Push(Scope::new("b").unwrap()));
         assert_ops_contain(&ops, &expected);
     }
@@ -900,11 +983,67 @@ mod tests {
 
         let syntax = syntax_set.find_syntax_by_extension("c").unwrap();
         let mut parse_state = ParseState::new(syntax);
-        let ops = parse_state.parse_line("c go_a a go_b b", &syntax_set);
+        let ops = parse_state.parse_line("c go_a a go_b b", &syntax_set).expect("#[cfg(test)]");
         let expected = (14, ScopeStackOp::Push(Scope::new("b").unwrap()));
         assert_ops_contain(&ops, &expected);
     }
 
+    #[test]
+    fn falls_back_to_plain_text_when_embedded_scope_is_missing() {
+        test_plain_text_fallback(r#"
+        name: Z
+        scope: source.z
+        file_extensions: [z]
+        contexts:
+          main:
+            - match: 'z'
+              scope: z
+            - match: 'go_x'
+              embed: scope:does.not.exist
+              escape: 'leave_x'
+        "#);
+    }
+
+    #[test]
+    fn falls_back_to_plain_text_when_embedded_file_is_missing() {
+        test_plain_text_fallback(r#"
+        name: Z
+        scope: source.z
+        file_extensions: [z]
+        contexts:
+          main:
+            - match: 'z'
+              scope: z
+            - match: 'go_x'
+              embed: DoesNotExist.sublime-syntax
+              escape: 'leave_x'
+        "#);
+    }
+
+    fn test_plain_text_fallback(syntax_definition: &str) {
+        let syntax =
+            SyntaxDefinition::load_from_str(syntax_definition, true, None).unwrap();
+
+        let mut builder = SyntaxSetBuilder::new();
+        builder.add_plain_text_syntax();
+        builder.add(syntax);
+        let syntax_set = builder.build();
+
+        let syntax = syntax_set.find_syntax_by_extension("z").unwrap();
+        let mut parse_state = ParseState::new(syntax);
+        let ops = parse_state.parse_line("z go_x x leave_x z", &syntax_set).unwrap();
+        let expected_ops = vec![
+            (0, ScopeStackOp::Push(Scope::new("source.z").unwrap())),
+            (0, ScopeStackOp::Push(Scope::new("z").unwrap())),
+            (1, ScopeStackOp::Pop(1)),
+            (6, ScopeStackOp::Push(Scope::new("text.plain").unwrap())),
+            (9, ScopeStackOp::Pop(1)),
+            (17, ScopeStackOp::Push(Scope::new("z").unwrap())),
+            (18, ScopeStackOp::Pop(1)),
+        ];
+        assert_eq!(ops, expected_ops);
+    }
+
     #[test]
     fn can_find_unlinked_contexts() {
         let syntax_set = {
@@ -925,7 +1064,7 @@ mod tests {
 
         let unlinked_contexts : Vec<String> = syntax_set.find_unlinked_contexts().into_iter().collect();
         assert_eq!(unlinked_contexts.len(), 1);
-        assert_eq!(unlinked_contexts[0], "Syntax 'A' with scope 'source.a' has unresolved context reference ByScope { scope: <source.b>, sub_context: Some(\"main\") }");
+        assert_eq!(unlinked_contexts[0], "Syntax 'A' with scope 'source.a' has unresolved context reference ByScope { scope: <source.b>, sub_context: Some(\"main\"), with_escape: false }");
     }
 
     #[test]
@@ -951,7 +1090,7 @@ mod tests {
             .map(|line| {
                 let syntax = syntax_set.find_syntax_by_extension("a").unwrap();
                 let mut parse_state = ParseState::new(syntax);
-                parse_state.parse_line(line, &syntax_set)
+                parse_state.parse_line(line, &syntax_set).expect("#[cfg(test)]")
             })
             .collect();
 
@@ -1013,13 +1152,13 @@ mod tests {
 
         let mut syntax = syntax_set.find_syntax_by_extension("a").unwrap();
         assert_eq!(syntax.name, "A improved");
-        syntax = syntax_set.find_syntax_by_scope(Scope::new(&"source.a").unwrap()).unwrap();
+        syntax = syntax_set.find_syntax_by_scope(Scope::new("source.a").unwrap()).unwrap();
         assert_eq!(syntax.name, "A improved");
-        syntax = syntax_set.find_syntax_by_first_line(&"syntax a").unwrap();
+        syntax = syntax_set.find_syntax_by_first_line("syntax a").unwrap();
         assert_eq!(syntax.name, "C");
 
         let mut parse_state = ParseState::new(syntax);
-        let ops = parse_state.parse_line("c go_a a", &syntax_set);
+        let ops = parse_state.parse_line("c go_a a", &syntax_set).expect("msg");
         let expected = (7, ScopeStackOp::Push(Scope::new("a2").unwrap()));
         assert_ops_contain(&ops, &expected);
     }
@@ -1034,7 +1173,7 @@ mod tests {
         let syntax = syntax_set.find_syntax_by_extension("yaml").unwrap();
 
         let mut parse_state = ParseState::new(syntax);
-        let ops = parse_state.parse_line("# test\n", &syntax_set);
+        let ops = parse_state.parse_line("# test\n", &syntax_set).expect("#[cfg(test)]");
         let expected = (0, ScopeStackOp::Push(Scope::new("comment.line.number-sign.yaml").unwrap()));
         assert_ops_contain(&ops, &expected);
     }
@@ -1106,12 +1245,12 @@ mod tests {
     }
 
     fn assert_prototype_only_on(expected: &[&str], syntax_set: &SyntaxSet, syntax: &SyntaxReference) {
-        for (name, id) in &syntax.contexts {
+        for (name, id) in syntax.context_ids() {
             if name == "__main" || name == "__start" {
                 // Skip special contexts
                 continue;
             }
-            let context = syntax_set.get_context(id);
+            let context = syntax_set.get_context(id).expect("#[cfg(test)]");
             if expected.contains(&name.as_str()) {
                 assert!(context.prototype.is_some(), "Expected context {} to have prototype", name);
             } else {
